@@ -1,45 +1,51 @@
 import asyncio
+import gc
 import json
 import logging
 import operator
 import re
 import sys
 from collections import defaultdict
+from itertools import product
+from textwrap import dedent
 from time import sleep
+from unittest import mock
 
 import cloudpickle
+import pytest
+from tlz import concat, first, frequencies, merge, valmap
+
 import dask
 from dask import delayed
-from tlz import merge, concat, valmap, first, frequencies
+from dask.compatibility import apply
 
-import pytest
-
-from distributed import Nanny, Worker, Client, wait, fire_and_forget
-from distributed.comm import Comm
-from distributed.core import connect, rpc, ConnectionPool, Status
-from distributed.scheduler import Scheduler
+from distributed import Client, Nanny, Worker, fire_and_forget, wait
 from distributed.client import wait
+from distributed.comm import Comm
+from distributed.compatibility import MACOS, WINDOWS
+from distributed.core import ConnectionPool, Status, connect, rpc
 from distributed.metrics import time
 from distributed.protocol.pickle import dumps
-from distributed.worker import dumps_function, dumps_task
-from distributed.utils import tmpfile, typename, TimeoutError
+from distributed.scheduler import MemoryState, Scheduler
+from distributed.utils import TimeoutError, tmpfile, typename
 from distributed.utils_test import (  # noqa: F401
     captured_logger,
     cleanup,
-    inc,
+    cluster,
     dec,
+    div,
     gen_cluster,
     gen_test,
-    slowinc,
+    inc,
+    loop,
+    nodebug,
     slowadd,
     slowdec,
-    cluster,
-    div,
-    varying,
+    slowinc,
     tls_only_security,
+    varying,
 )
-from distributed.utils_test import loop, nodebug  # noqa: F401
-from dask.compatibility import apply
+from distributed.worker import dumps_function, dumps_task
 
 if sys.version_info < (3, 8):
     try:
@@ -407,6 +413,7 @@ async def test_delete_data(c, s, a, b):
 async def test_delete(c, s, a):
     x = c.submit(inc, 1)
     await x
+    assert x.key in s.tasks
     assert x.key in a.data
 
     await c._cancel(x)
@@ -416,6 +423,10 @@ async def test_delete(c, s, a):
         await asyncio.sleep(0.01)
         assert time() < start + 5
 
+    assert x.key not in s.tasks
+
+    s.report_on_key(key=x.key)
+
 
 @gen_cluster()
 async def test_filtered_communication(s, a, b):
@@ -1039,11 +1050,12 @@ async def test_no_workers_to_memory(c, s):
 
 @gen_cluster(client=True)
 async def test_no_worker_to_memory_restrictions(c, s, a, b):
-    x = delayed(slowinc)(1, delay=0.4)
-    y = delayed(slowinc)(x, delay=0.4)
-    z = delayed(slowinc)(y, delay=0.4)
+    with dask.annotate(workers="alice"):
+        x = delayed(slowinc)(1, delay=0.4)
+        y = delayed(slowinc)(x, delay=0.4)
+        z = delayed(slowinc)(y, delay=0.4)
 
-    yy, zz = c.persist([y, z], workers={(x, y, z): "alice"})
+    yy, zz = c.persist([y, z], optimize_graph=False)
 
     while not s.tasks:
         await asyncio.sleep(0.01)
@@ -1088,6 +1100,15 @@ async def test_run_on_scheduler(c, s, a, b):
     assert response == s.address
 
 
+@gen_cluster(client=True, config={"distributed.scheduler.pickle": False})
+async def test_run_on_scheduler_disabled(c, s, a, b):
+    def f(dask_scheduler=None):
+        return dask_scheduler.address
+
+    with pytest.raises(ValueError, match="disallowed from deserializing"):
+        await c._run_on_scheduler(f)
+
+
 @gen_cluster(client=True)
 async def test_close_worker(c, s, a, b):
     assert len(s.workers) == 2
@@ -1103,7 +1124,7 @@ async def test_close_worker(c, s, a, b):
 
 
 @pytest.mark.slow
-@gen_cluster(client=True, Worker=Nanny, timeout=20)
+@gen_cluster(client=True, Worker=Nanny)
 async def test_close_nanny(c, s, a, b):
     assert len(s.workers) == 2
 
@@ -1134,7 +1155,7 @@ async def test_close_nanny(c, s, a, b):
         assert time() < start + 10
 
 
-@gen_cluster(client=True, timeout=20)
+@gen_cluster(client=True)
 async def test_retire_workers_close(c, s, a, b):
     await s.retire_workers(close_workers=True)
     assert not s.workers
@@ -1142,7 +1163,7 @@ async def test_retire_workers_close(c, s, a, b):
         await asyncio.sleep(0.01)
 
 
-@gen_cluster(client=True, timeout=20, Worker=Nanny)
+@gen_cluster(client=True, Worker=Nanny)
 async def test_retire_nannies_close(c, s, a, b):
     nannies = [a, b]
     await s.retire_workers(close_workers=True, remove=True)
@@ -1182,7 +1203,7 @@ async def test_scheduler_file():
         await s.close()
 
 
-@pytest.mark.xfail(reason="")
+@pytest.mark.xfail()
 @gen_cluster(client=True, nthreads=[])
 async def test_non_existent_worker(c, s):
     with dask.config.set({"distributed.comm.timeouts.connect": "100ms"}):
@@ -1417,9 +1438,11 @@ async def test_retries(c, s, a, b):
     exc_info.match("one")
 
 
-@pytest.mark.xfail(reason="second worker also errant for some reason")
-@gen_cluster(client=True, nthreads=[("127.0.0.1", 1)] * 3, timeout=5)
-async def test_mising_data_errant_worker(c, s, w1, w2, w3):
+@pytest.mark.flaky(
+    reruns=10, reruns_delay=5, reason="second worker also errant for some reason"
+)
+@gen_cluster(client=True, nthreads=[("127.0.0.1", 1)] * 3)
+async def test_missing_data_errant_worker(c, s, w1, w2, w3):
     with dask.config.set({"distributed.comm.timeouts.connect": "1s"}):
         np = pytest.importorskip("numpy")
 
@@ -1607,6 +1630,7 @@ async def test_collect_versions(c, s, a, b):
     assert cs.versions == w1.versions == w2.versions
 
 
+@pytest.mark.xfail(reason="flaky and re-fails on rerun")
 @gen_cluster(client=True, config={"distributed.scheduler.idle-timeout": "500ms"})
 async def test_idle_timeout(c, s, a, b):
     beginning = time()
@@ -1713,6 +1737,18 @@ async def test_dashboard_address():
     assert s.services["dashboard"].port
     await s.close()
 
+    s = await Scheduler(dashboard_address="127.0.0.1:8901,127.0.0.1:8902", port=0)
+    assert s.services["dashboard"].port == 8901
+    await s.close()
+
+    s = await Scheduler(dashboard_address=":8901,:8902", port=0)
+    assert s.services["dashboard"].port == 8901
+    await s.close()
+
+    s = await Scheduler(dashboard_address=[8901, 8902], port=0)
+    assert s.services["dashboard"].port == 8901
+    await s.close()
+
 
 @gen_cluster(client=True)
 async def test_adaptive_target(c, s, a, b):
@@ -1813,9 +1849,6 @@ async def test_get_task_duration(c, s, a, b):
 
 
 @pytest.mark.asyncio
-@pytest.mark.skipif(
-    sys.version_info < (3, 7), reason="asyncio.all_tasks not implemented"
-)
 async def test_no_danglng_asyncio_tasks(cleanup):
     start = asyncio.all_tasks()
     async with Scheduler(port=0) as s:
@@ -1827,12 +1860,30 @@ async def test_no_danglng_asyncio_tasks(cleanup):
     assert tasks == start
 
 
-@gen_cluster(client=True)
+class NoSchedulerDelayWorker(Worker):
+    """Custom worker class which does not update `scheduler_delay`.
+
+    This worker class is useful for some tests which make time
+    comparisons using times reported from workers.
+    """
+
+    @property
+    def scheduler_delay(self):
+        return 0
+
+    @scheduler_delay.setter
+    def scheduler_delay(self, value):
+        pass
+
+
+@gen_cluster(client=True, Worker=NoSchedulerDelayWorker)
 async def test_task_groups(c, s, a, b):
+    start = time()
     da = pytest.importorskip("dask.array")
     x = da.arange(100, chunks=(20,))
     y = (x + 1).persist(optimize_graph=False)
     y = await y
+    stop = time()
 
     tg = s.task_groups[x.name]
     tp = s.task_prefixes["arange"]
@@ -1867,6 +1918,9 @@ async def test_task_groups(c, s, a, b):
     assert tg.states["forgotten"] == 5
     # Ensure TaskGroup is removed once all tasks are in forgotten state
     assert tg.name not in s.task_groups
+    assert tg.start > start
+    assert tg.stop < stop
+    assert "compute" in tg.all_durations
     assert sys.getrefcount(tg) == 2
 
 
@@ -2021,17 +2075,18 @@ async def test_gather_allow_worker_reconnect(c, s, a, b):
     its results instead of recomputing them.
     """
     # GH3246
-    ALREADY_CALCULATED = []
+    already_calculated = []
 
     import time
 
     def inc_slow(x):
         # Once the graph below is rescheduled this computation runs again. We
         # need to sleep for at least 0.5 seconds to give the worker a chance to
-        # reconnect (Heartbeat timing)
-        if x in ALREADY_CALCULATED:
-            time.sleep(1)
-        ALREADY_CALCULATED.append(x)
+        # reconnect (Heartbeat timing). In slow CI situations, the actual
+        # reconnect might take a bit longer, therefore wait more
+        if x in already_calculated:
+            time.sleep(2)
+        already_calculated.append(x)
         return x + 1
 
     x = c.submit(inc_slow, 1)
@@ -2042,18 +2097,16 @@ async def test_gather_allow_worker_reconnect(c, s, a, b):
 
     z = c.submit(reducer, x, y)
 
-    s.rpc = await FlakyConnectionPool(failing_connections=4)
+    s.rpc = await FlakyConnectionPool(failing_connections=1)
 
-    with dask.config.set(
-        {"distributed.comm.retry.delay_min": 0.5, "distributed.comm.retry.count": 3}
-    ):
+    # This behaviour is independent of retries. Remove them to reduce complexity
+    # of this setup
+    with dask.config.set({"distributed.comm.retry.count": 0}):
         with captured_logger(
             logging.getLogger("distributed.scheduler")
         ) as sched_logger, captured_logger(
             logging.getLogger("distributed.client")
-        ) as client_logger, captured_logger(
-            logging.getLogger("distributed.utils_comm")
-        ) as utils_comm_logger:
+        ) as client_logger:
             # Gather using the client (as an ordinary user would)
             # Upon a missing key, the client will reschedule the computations
             res = await c.gather(z)
@@ -2062,14 +2115,13 @@ async def test_gather_allow_worker_reconnect(c, s, a, b):
 
     sched_logger = sched_logger.getvalue()
     client_logger = client_logger.getvalue()
-    utils_comm_logger = utils_comm_logger.getvalue()
 
-    # Ensure that the communication was done via the scheduler, i.e. we actually hit a bad connection
+    # Ensure that the communication was done via the scheduler, i.e. we actually hit a
+    # bad connection
     assert s.rpc.cnn_count > 0
 
-    assert "Retrying get_data_from_worker after exception" in utils_comm_logger
-
-    # The reducer task was actually not found upon first collection. The client will reschedule the graph
+    # The reducer task was actually not found upon first collection. The client will
+    # reschedule the graph
     assert "Couldn't gather 1 keys, rescheduling" in client_logger
     # There will also be a `Unexpected worker completed task` message but this
     # is rather an artifact and not the intention
@@ -2086,7 +2138,6 @@ async def test_gather_allow_worker_reconnect(c, s, a, b):
     ]
     assert len(transitions_to_processing) == 1
 
-    starts = []
     finish_processing_transitions = 0
     for transition in s.transition_log:
         key, start, finish, recommendations, timestamp = transition
@@ -2167,7 +2218,7 @@ async def test_unknown_task_duration_config(s, a, b):
     assert s.idle_since == s.time_started
 
 
-@gen_cluster(client=True, timeout=1000)
+@gen_cluster(client=True, timeout=None)
 async def test_retire_state_change(c, s, a, b):
     np = pytest.importorskip("numpy")
     y = c.map(lambda x: x ** 2, range(10))
@@ -2179,3 +2230,278 @@ async def test_retire_state_change(c, s, a, b):
         step = c.compute(foo)
         c.gather(step)
     await c.retire_workers(workers=[a.address])
+
+
+@gen_cluster(client=True, config={"distributed.scheduler.events-log-length": 3})
+async def test_configurable_events_log_length(c, s, a, b):
+    s.log_event("test", "dummy message 1")
+    assert len(s.events["test"]) == 1
+    s.log_event("test", "dummy message 2")
+    s.log_event("test", "dummy message 3")
+    assert len(s.events["test"]) == 3
+
+    # adding a forth message will drop the first one and length stays at 3
+    s.log_event("test", "dummy message 4")
+    assert len(s.events["test"]) == 3
+    assert s.events["test"][0][1] == "dummy message 2"
+    assert s.events["test"][1][1] == "dummy message 3"
+    assert s.events["test"][2][1] == "dummy message 4"
+
+
+@gen_cluster()
+async def test_get_worker_monitor_info(s, a, b):
+    res = await s.get_worker_monitor_info()
+    ms = ["cpu", "time", "read_bytes", "write_bytes"]
+    if not WINDOWS:
+        ms += ["num_fds"]
+    for w in (a, b):
+        assert all(res[w.address]["range_query"][m] is not None for m in ms)
+        assert res[w.address]["count"] is not None
+        assert res[w.address]["last_time"] is not None
+
+
+@gen_cluster(client=True)
+async def test_quiet_cluster_round_robin(c, s, a, b):
+    await c.submit(inc, 1)
+    await c.submit(inc, 2)
+    await c.submit(inc, 3)
+    assert a.log and b.log
+
+
+def test_memorystate():
+    m = MemoryState(process=100, unmanaged_old=15, managed=80, managed_spilled=12)
+    assert m.process == 100
+    assert m.managed == 80
+    assert m.managed_in_memory == 68
+    assert m.managed_spilled == 12
+    assert m.unmanaged == 32
+    assert m.unmanaged_old == 15
+    assert m.unmanaged_recent == 17
+    assert m.optimistic == 83
+
+    assert (
+        repr(m)
+        == dedent(
+            """
+            Managed by Dask       : 80 B
+              - in process memory : 68 B
+              - spilled to disk   : 12 B
+            Process memory (RSS)  : 100 B
+              - managed by Dask   : 68 B
+              - unmanaged (old)   : 15 B
+              - unmanaged (recent): 17 B
+            """
+        ).lstrip()
+    )
+
+
+def test_memorystate_sum():
+    m1 = MemoryState(process=100, unmanaged_old=15, managed=80, managed_spilled=12)
+    m2 = MemoryState(process=80, unmanaged_old=10, managed=60, managed_spilled=2)
+    m3 = MemoryState.sum(m1, m2)
+    assert m3.process == 180
+    assert m3.unmanaged_old == 25
+    assert m3.managed == 140
+    assert m3.managed_spilled == 14
+
+
+@pytest.mark.parametrize(
+    "process,unmanaged_old,managed,managed_spilled", list(product(*[[0, 1, 2, 3]] * 4))
+)
+def test_memorystate_adds_up(process, unmanaged_old, managed, managed_spilled):
+    """Input data is massaged by __init__ so that everything adds up by construction"""
+    m = MemoryState(
+        process=process,
+        unmanaged_old=unmanaged_old,
+        managed=managed,
+        managed_spilled=managed_spilled,
+    )
+    assert m.managed_in_memory + m.unmanaged == m.process
+    assert m.managed_in_memory + m.managed_spilled == m.managed
+    assert m.unmanaged_old + m.unmanaged_recent == m.unmanaged
+    assert m.optimistic + m.unmanaged_recent == m.process
+
+
+def leaking(out_mib, leak_mib, sleep_time):
+    if leak_mib:
+        global __test_leak
+        __test_leak = "x" * (leak_mib * 2 ** 20)
+    out = "x" * (out_mib * 2 ** 20)
+    sleep(sleep_time)
+    return out
+
+
+def clear_leak():
+    global __test_leak
+    del __test_leak
+    gc.collect()
+
+
+async def assert_memory(scheduler_or_workerstate, attr: str, min_, max_, timeout=10):
+    t0 = time()
+    while True:
+        minfo = scheduler_or_workerstate.memory
+        nbytes = getattr(minfo, attr)
+        if min_ * 2 ** 20 <= nbytes <= max_ * 2 ** 20:
+            return
+        if time() - t0 > timeout:
+            raise TimeoutError(
+                f"Expected {min_} MiB <= {attr} <= {max_} MiB; got:\n{minfo!r}"
+            )
+        await asyncio.sleep(0.1)
+
+
+# This test is heavily influenced by hard-to-control factors such as memory management
+# by the Python interpreter and the OS, so it occasionally glitches
+@pytest.mark.flaky(reruns=3, reruns_delay=5)
+# ~33s runtime, or distributed.memory.recent_to_old_time + 3s
+@pytest.mark.slow
+@gen_cluster(
+    client=True, Worker=Nanny, worker_kwargs={"memory_limit": "500 MiB"}, timeout=60
+)
+async def test_memory(c, s, *_):
+    pytest.importorskip("zict")
+
+    # WorkerState objects, as opposed to the Nanny objects passed by gen_cluster
+    a, b = s.workers.values()
+
+    s_m0 = s.memory
+    assert s_m0.process == a.memory.process + b.memory.process
+    assert s_m0.managed == 0
+    assert a.memory.managed == 0
+    assert b.memory.managed == 0
+    # When a worker first goes online, its RAM is immediately counted as
+    # unmanaged_old
+    await assert_memory(s, "unmanaged_recent", 0, 40, timeout=0)
+    await assert_memory(a, "unmanaged_recent", 0, 20, timeout=0)
+    await assert_memory(b, "unmanaged_recent", 0, 20, timeout=0)
+
+    f1 = c.submit(leaking, 100, 50, 5, pure=False, workers=[a.name])
+    f2 = c.submit(leaking, 100, 50, 5, pure=False, workers=[b.name])
+    await assert_memory(s, "unmanaged_recent", 300, 380)
+    await assert_memory(a, "unmanaged_recent", 150, 190)
+    await assert_memory(b, "unmanaged_recent", 150, 190)
+    await wait([f1, f2])
+
+    # On each worker, we now have 100 MiB managed + 50 MiB fresh leak
+    await assert_memory(s, "managed_in_memory", 200, 201)
+    await assert_memory(a, "managed_in_memory", 100, 101)
+    await assert_memory(b, "managed_in_memory", 100, 101)
+    await assert_memory(s, "unmanaged_recent", 100, 180)
+    await assert_memory(a, "unmanaged_recent", 50, 90)
+    await assert_memory(b, "unmanaged_recent", 50, 90)
+
+    # Force the output of f1 and f2 to spill to disk.
+    # With target=0.6 and memory_limit=500 MiB, we'll start spilling at 300 MiB
+    # process memory per worker, or roughly after 3~7 rounds of the below depending
+    # on how much RAM the interpreter is using.
+    more_futs = []
+    for _ in range(8):
+        if s.memory.managed_spilled > 0:
+            break
+        more_futs += [
+            c.submit(leaking, 20, 0, 0, pure=False, workers=[a.name]),
+            c.submit(leaking, 20, 0, 0, pure=False, workers=[b.name]),
+        ]
+        await asyncio.sleep(2)
+    await assert_memory(s, "managed_spilled", 1, 999)
+
+    # Wait for the spilling to finish. Note that this does not make the test take
+    # longer as we're waiting for recent_to_old_time anyway.
+    await asyncio.sleep(10)
+
+    # Delete spilled keys
+    prev = s.memory
+    del f1
+    del f2
+    await assert_memory(s, "managed_spilled", 0, prev.managed_spilled / 2 ** 20 - 19)
+
+    # Empty the cluster, with the exception of leaked memory
+    del more_futs
+    await assert_memory(s, "managed", 0, 0)
+
+    orig_unmanaged = s_m0.unmanaged / 2 ** 20
+    orig_old = s_m0.unmanaged_old / 2 ** 20
+
+    # Wait until 30s have passed since the spill to observe unmanaged_recent
+    # transition into unmanaged_old
+    await c.run(gc.collect)
+    await assert_memory(s, "unmanaged_recent", 0, 90, timeout=40)
+    await assert_memory(
+        s,
+        "unmanaged_old",
+        orig_old + 90,
+        # On MacOS, the process memory of the Python interpreter does not shrink as
+        # fast as on Linux/Windows
+        9999 if MACOS else orig_old + 190,
+        timeout=40,
+    )
+
+    # When the leaked memory is cleared, unmanaged and unmanaged_old drop
+    # On MacOS, the process memory of the Python interpreter does not shrink as fast
+    # as on Linux/Windows
+    if not MACOS:
+        await c.run(clear_leak)
+        await assert_memory(s, "unmanaged", 0, orig_unmanaged + 95)
+        await assert_memory(s, "unmanaged_old", 0, orig_old + 95)
+        await assert_memory(s, "unmanaged_recent", 0, 90)
+
+
+@gen_cluster(client=True, worker_kwargs={"memory_limit": 0})
+async def test_memory_no_zict(c, s, a, b):
+    """When Worker.data is not a SpillBuffer, test that querying managed_spilled
+    defaults to 0 and doesn't raise KeyError
+    """
+    await c.wait_for_workers(2)
+    assert isinstance(a.data, dict)
+    assert isinstance(b.data, dict)
+    f = c.submit(leaking, 10, 0, 0)
+    await f
+    assert 10 * 2 ** 20 < s.memory.managed_in_memory < 11 * 2 ** 20
+    assert s.memory.managed_spilled == 0
+
+
+@gen_cluster(nthreads=[])
+async def test_memory_no_workers(s):
+    assert s.memory.process == 0
+    assert s.memory.managed == 0
+
+
+@gen_cluster(client=True, nthreads=[])
+async def test_memory_is_none(c, s):
+    """If Worker.heartbeat() runs before Worker.monitor.update(), then
+    Worker.metrics["memory"] will be None and will need special handling in
+    Worker.memory and Scheduler.heartbeat_worker().
+    """
+    with mock.patch("distributed.system_monitor.SystemMonitor.update"):
+        async with Worker(s.address, nthreads=1) as w:
+            await c.wait_for_workers(1)
+            f = await c.scatter(123)
+            await w.heartbeat()
+            assert s.memory.process == 0  # Forced from None
+            assert s.memory.managed == 0  # Capped by process even if we do have keys
+            assert s.memory.managed_in_memory == 0
+            assert s.memory.managed_spilled == 0
+            assert s.memory.unmanaged == 0
+            assert s.memory.unmanaged_old == 0
+            assert s.memory.unmanaged_recent == 0
+
+
+@gen_cluster()
+async def test_close_scheduler__close_workers_Worker(s, a, b):
+    with captured_logger("distributed.comm", level=logging.DEBUG) as log:
+        await s.close(close_workers=True)
+        while not a.status == Status.closed:
+            await asyncio.sleep(0.05)
+    log = log.getvalue()
+    assert "retry" not in log
+
+
+@gen_cluster(Worker=Nanny)
+async def test_close_scheduler__close_workers_Nanny(s, a, b):
+    with captured_logger("distributed.comm", level=logging.DEBUG) as log:
+        await s.close(close_workers=True)
+        while not a.status == Status.closed:
+            await asyncio.sleep(0.05)
+    log = log.getvalue()
+    assert "retry" not in log
